#include "atomicOperations.h"
#include "samax_macros.h"

// 
// Generic vector max kernel.
//
// - Designed for architectures with atomicCAS, i.e., sm_11 and higher.
// - Simple design for all capabilities
// - Handled by splitting matrix as
// --------------------------
// |                 |      |
// |                 |      |
// |                 |      |
// | left tiled part | rest | 
// |                 |      |
// |                 |      |
// |                 |      |
// --------------------------
// - Kernel is generic and works for all cases
// - Unrolling of inner loops
//     1) to reduce instruction count from loop
//     2) to collect shared loads into 128 bit transactions (compiler should do that)
//   * Level determined by WORKSIZE_n
//   * For WORKSIZE_n = 1:
//       Given by a constant: full unroll
//       Given by a variable: unroll level 4

template <int BLOCKSIZE_n>
inline __device__ void block_reduce_max(float *y, volatile float *smem, float sum, int tid)
{
	smem[tid] = sum;
	__syncthreads();
	if (BLOCKSIZE_n >= 1024) { if (tid < 512) { smem[tid] = max(smem[tid], smem[tid + 512]); } __syncthreads(); }
	if (BLOCKSIZE_n >= 512) { if (tid < 256) { smem[tid] = max(smem[tid], smem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 480) { if (tid < 224) { smem[tid] = max(smem[tid], smem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 448) { if (tid < 192) { smem[tid] = max(smem[tid], smem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 416) { if (tid < 160) { smem[tid] = max(smem[tid], smem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 384) { if (tid < 128) { smem[tid] = max(smem[tid], smem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n >= 256) { if (tid < 128) { smem[tid] = max(smem[tid], smem[tid + 128]); } __syncthreads(); }
	if (BLOCKSIZE_n == 352) { if (tid < 96) { smem[tid] = max(smem[tid], smem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 224) { if (tid < 96) { smem[tid] = max(smem[tid], smem[tid + 128]); } __syncthreads(); }
	if (BLOCKSIZE_n == 320) { if (tid < 64) { smem[tid] = max(smem[tid], smem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 192) { if (tid < 64) { smem[tid] = max(smem[tid], smem[tid + 128]); } __syncthreads(); }
	if (BLOCKSIZE_n >= 128) { if (tid < 64) { smem[tid] = max(smem[tid], smem[tid + 64]); } __syncthreads(); }
	if (BLOCKSIZE_n == 288) { if (tid < 32) { smem[tid] = max(smem[tid], smem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 160) { if (tid < 32) { smem[tid] = max(smem[tid], smem[tid + 128]); } __syncthreads(); }
	if (BLOCKSIZE_n == 96) { if (tid < 32) { smem[tid] = max(smem[tid], smem[tid + 64]); } __syncthreads(); }
	if (BLOCKSIZE_n >= 64) { if (tid < 32) { smem[tid] = max(smem[tid], smem[tid + 32]); } }
	if (BLOCKSIZE_n >= 32) { if (tid < 16) { smem[tid] = max(smem[tid], smem[tid + 16]); } }
	if (BLOCKSIZE_n >= 16) { if (tid < 8) { smem[tid] = max(smem[tid], smem[tid + 8]); } }
	if (BLOCKSIZE_n >= 8) { if (tid < 4) { smem[tid] = max(smem[tid], smem[tid + 4]); } }
	if (BLOCKSIZE_n >= 4) { if (tid < 2) { smem[tid] = max(smem[tid], smem[tid + 2]); } }
	if (BLOCKSIZE_n >= 2) { if (tid < 1) { smem[tid] = max(smem[tid], smem[tid + 1]); } }
	if (tid == 0) { atomicMax(y, smem[tid]); }
}

template <int BLOCKSIZE_n, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_n, MIN_BLOCKS_PER_MP)
__global__ void samax_kernel(const float *x, float *y, int n, int ntile)
{
	const int tid = threadIdx.x;
	const int idx = blockIdx.x * BLOCKSIZE_n * WORKSIZE_n;
	x += idx + tid;
	float sum = x[0];
	if (idx != ntile)
	{
        #pragma unroll
		inner_loop(WORKSIZE_n);
	}
	else
	{
        #pragma unroll 4
		inner_loop2(n - idx - tid);
	}
	__shared__ volatile float smem[BLOCKSIZE_n];
	block_reduce_max<BLOCKSIZE_n>(y, smem, sum, tid);
}

