#include "atomicAdd.h"
#include "sgemv_col_major_macros.h"

// 
// Generic matrix-vector multiplication kernel.
//
// - Designed for architectures with atomicCAS, i.e., sm_11 and higher.
// - Split into 4 separate device kernels.
// -----------------
// |         |     |
// |    1    |  2  |
// |         |     |
// |---------------|
// |    3    |  4  |
// -----------------
// - Valid for all cases.
// - Uses atomicCAS (sm_11-sm_13) or atomicAdd (>=sm_20) for grid level reduction.
// - Unrolling of inner loops
//     * Given by a constant: full unroll
//     * Given by a variable: unroll level 4
//     1) to reduce instruction count from loop
//     2) to collect shared loads unsigned into 128 bit transactions (compiler should do that)
//

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void kernel1(float alpha, const float* A, int lda, const float* x, float* y, float *x_s, int m, int tid)
{
	//
	// The top-left fully tiled part of the Matrix
	//
	initialize_sums(WORKSIZE_m);
	for (int j = 0; j < BLOCKSIZE_m * WORKSIZE_n; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
        #pragma unroll
		inner_loop(BLOCKSIZE_m);
	}
	atomicAdd_sums(BLOCKSIZE_m);
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void kernel2(float alpha, const float* A, int lda, const float* x, float* y, float *x_s, int m, int nrest, int tid)
{
	//
	// The top-right m-tiled part of the Matrix
	//
	int ntile = (nrest / BLOCKSIZE_m) * BLOCKSIZE_m;
	initialize_sums(WORKSIZE_m);
	for (int j = 0; j < ntile; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
        #pragma unroll
		inner_loop(BLOCKSIZE_m);
	}
	__syncthreads();
	if (tid < nrest - ntile)
		x_s[tid]  = x[ntile];
	__syncthreads();
    #pragma unroll 4
	inner_loop(nrest - ntile);
	atomicAdd_sums(BLOCKSIZE_m);
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void kernel3(float alpha, const float* A, int lda, const float* x, float* y, float *x_s, int m, int mrest, int tid)
{
	//
	// The bottom-left n-tiled part of the Matrix
	//
	initialize_sums(WORKSIZE_m);
	for (int j = 0; j < BLOCKSIZE_m * WORKSIZE_n; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
		if (tid < mrest)
		{
            #pragma unroll
			inner_loop_check_bounds(tid, mrest, BLOCKSIZE_m);
		}
	}
	if (tid < mrest)
	{
		atomicAdd_sums_check_bounds(tid, mrest, BLOCKSIZE_m);
	}
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void kernel4(float alpha, const float* A, int lda, const float* x, float* y, float *x_s, int m, int mrest, int nrest, int tid)
{
	//
	// The bottom-right rest of the Matrix
	//
	int ntile = (nrest / BLOCKSIZE_m) * BLOCKSIZE_m;
	initialize_sums(WORKSIZE_m);
	for (int j = 0; j < ntile; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
		if (tid < mrest)
		{
            #pragma unroll
			inner_loop_check_bounds(tid, mrest, BLOCKSIZE_m);
		}
	}
	__syncthreads();
	if (tid < nrest - ntile)
		x_s[tid]  = x[ntile];
	__syncthreads();
	if (tid < mrest)
	{
        #pragma unroll 4
		inner_loop_check_bounds(tid, mrest, nrest - ntile);
		atomicAdd_sums_check_bounds(tid, mrest, BLOCKSIZE_m);
	}
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void sgemv_generic_kernel(float alpha, const float* A, int lda, const float* x, float* y, int m, int n, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = blockIdx.x * BLOCKSIZE_m * WORKSIZE_m;
	int col = blockIdx.y * BLOCKSIZE_m * WORKSIZE_n;
	__shared__ float x_s[BLOCKSIZE_m];
	A += tid + row + col * lda;
	x += tid + col;
	y += tid + row;
	if (row !=  mtile)
	{
		if (col != ntile)
		{	
			kernel1<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(alpha, A, lda, x, y, x_s, m, tid);
		}
		else
		{
			kernel2<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(alpha, A, lda, x, y, x_s, m, n - ntile, tid);
		}
	}
	else
	{
		if (col != ntile)
		{
			kernel3<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(alpha, A, lda, x, y, x_s, m, m - mtile, tid);
		}
		else
		{
			kernel4<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(alpha, A, lda, x, y, x_s, m, m - mtile, n - ntile, tid);
		}
	}
}

//
// Typical matrix-vector multiplication kernel: for m > 100 * n (tall and not too wide) cases.
//
// - Simple design for all capabilities
// - Handled by splitting matrix as
// --------------------------
// |                 |      |
// |                 |      |
// |                 |      |
// | left tiled part | rest | 
// |                 |      |
// |                 |      |
// |                 |      |
// --------------------------
// - Kernel is generic and works for all cases
// - Unrolling of inner loops
//     1) to reduce instruction count from loop
//     2) to collect shared loads unsigned into 128 bit transactions (compiler should do that)
//   * Level determined by WORKSIZE_n
//   * For WORKSIZE_n = 1:
//       Given by a constant: full unroll
//       Given by a variable: unroll level 4

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m)
__global__ void sgemv_one_thread_per_row_kernel(float alpha, const float* A, int lda, const float* x, float* y, int m, int n, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = tid + blockIdx.x * BLOCKSIZE_m * WORKSIZE_m;
	__shared__ float x_s[BLOCKSIZE_m];
	A += row;
	x += tid;
	y += row;
	initialize_sums(WORKSIZE_m);
	// Left tiled part of the row
	for (int j = 0; j < ntile; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
		if (row < m)
		{
			// Control unroll level by template parameter (CUDA __Pragma doesn't work properly for macros)
			if (WORKSIZE_n == 1)
			{	// NB: full unroll
				#pragma unroll
				inner_loop_check_bounds(row, m, BLOCKSIZE_m);
			}
			if (WORKSIZE_n == 2)
			{					
				#pragma unroll 2
				inner_loop_check_bounds(row, m, BLOCKSIZE_m);
			}
			if (WORKSIZE_n == 3)
			{					
				#pragma unroll 3
				inner_loop_check_bounds(row, m, BLOCKSIZE_m);
			}
			if (WORKSIZE_n == 4)
			{					
				#pragma unroll 4
				inner_loop_check_bounds(row, m, BLOCKSIZE_m);
			}
			if (WORKSIZE_n == 5)
			{					
				#pragma unroll 5
				inner_loop_check_bounds(row, m, BLOCKSIZE_m);
			}
			if (WORKSIZE_n == 6)
			{					
				#pragma unroll 6
				inner_loop_check_bounds(row, m, BLOCKSIZE_m);
			}
			if (WORKSIZE_n == 7)
			{					
				#pragma unroll 7
				inner_loop_check_bounds(row, m, BLOCKSIZE_m);
			}
			if (WORKSIZE_n == 8)
			{					
				#pragma unroll 8
				inner_loop_check_bounds(row, m, BLOCKSIZE_m);
			}
		}
	}
	// Right rest of row
	__syncthreads();
	if (tid < n - ntile)
		x_s[tid]  = x[ntile];
	__syncthreads();
	if (row < m)
	{
		// Control unroll level by template parameter (CUDA __Pragma doesn't work properly for macros)
		if (WORKSIZE_n == 1)
		{	// NB: unroll 4, see kernel description
            #pragma unroll 4
			inner_loop_check_bounds(row, m, n - ntile);
		}
		if (WORKSIZE_n == 2)
		{					
            #pragma unroll 2
			inner_loop_check_bounds(row, m, n - ntile);
		}
		if (WORKSIZE_n == 3)
		{					
            #pragma unroll 3
			inner_loop_check_bounds(row, m, n - ntile);
		}
		if (WORKSIZE_n == 4)
		{					
            #pragma unroll 4
			inner_loop_check_bounds(row, m, n - ntile);
		}
		if (WORKSIZE_n == 5)
		{					
            #pragma unroll 5
			inner_loop_check_bounds(row, m, n - ntile);
		}
		if (WORKSIZE_n == 6)
		{					
            #pragma unroll 6
			inner_loop_check_bounds(row, m, n - ntile);
		}
		if (WORKSIZE_n == 7)
		{					
            #pragma unroll 7
			inner_loop_check_bounds(row, m, n - ntile);
		}
		if (WORKSIZE_n == 8)
		{					
            #pragma unroll 8
			inner_loop_check_bounds(row, m, n - ntile);
		}
		atomicAdd_sums_check_bounds_ts(row, m, BLOCKSIZE_m);
	}
}

// 
// Several threads per row matrix-vector multiplication kernel.
//
// - Designed for architectures with atomicCAS, i.e., sm_11 and higher.
// - Handled by splitting matrix as
// --------------------------
// |                 |      |
// |                 |      |
// |                 |      |
// | left tiled part | rest | 
// |                 |      |
// |                 |      |
// |                 |      |
// --------------------------
// - Valid for all cases.
// - Uses atomicCAS (sm_11-sm_13) or atomicAdd (>=sm_20) for grid level reduction.
// - Unrolling of inner loops
//     * Given by a constant: full unroll
//     * Given by a variable: unroll level 4
//     1) to reduce instruction count from loop
//     2) to collect shared loads unsigned into 128 bit transactions (compiler should do that)
//

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void left(float alpha, const float* A, int lda, const float* x, float* y, float *x_s, int m, int tid)
{
	//
	// The top-left fully tiled part of the Matrix
	//
	float sum = 0;
	for (int j = 0; j < BLOCKSIZE_m * WORKSIZE_n; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
		if (WORKSIZE_m == 1)
		{	// NB: full unroll
            #pragma unroll
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 2)
		{					
			#pragma unroll 2
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 3)
		{					
			#pragma unroll 3
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 4)
		{					
			#pragma unroll 4
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 5)
		{					
			#pragma unroll 5
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 6)
		{					
			#pragma unroll 6
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 7)
		{					
			#pragma unroll 7
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 8)
		{					
			#pragma unroll 8
			inner_loop_g2(BLOCKSIZE_m);
		}
	}
	atomicAdd(y, alpha * sum);
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void rest(float alpha, const float* A, int lda, const float* x, float* y, float *x_s, int m, int nrest, int tid)
{
	//
	// The top-right m-tiled part of the Matrix
	//
	int ntile = (nrest / BLOCKSIZE_m) * BLOCKSIZE_m;
	float sum = 0;
	for (int j = 0; j < ntile; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
		if (WORKSIZE_m == 1)
		{	// NB: full unroll
            #pragma unroll
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 2)
		{					
			#pragma unroll 2
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 3)
		{					
			#pragma unroll 3
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 4)
		{					
			#pragma unroll 4
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 5)
		{					
			#pragma unroll 5
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 6)
		{					
			#pragma unroll 6
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 7)
		{					
			#pragma unroll 7
			inner_loop_g2(BLOCKSIZE_m);
		}
		if (WORKSIZE_m == 8)
		{					
			#pragma unroll 8
			inner_loop_g2(BLOCKSIZE_m);
		}
	}
	__syncthreads();
	if (tid < nrest - ntile)
		x_s[tid]  = x[ntile];
	__syncthreads();
    #pragma unroll 4
	inner_loop_g2(nrest - ntile);
	atomicAdd(y, alpha * sum);
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void sgemv_several_threads_per_row_kernel(float alpha, const float* A, int lda, const float* x, float* y, int m, int n, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = blockIdx.x * BLOCKSIZE_m;
	int col = blockIdx.y * BLOCKSIZE_m * WORKSIZE_n;
	__shared__ float x_s[BLOCKSIZE_m];
	A += tid + row + col * lda;
	x += tid + col;
	y += tid + row;
	if (col != ntile)
	{	
		left<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(alpha, A, lda, x, y, x_s, m, tid);
	}
	else
	{
		rest<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(alpha, A, lda, x, y, x_s, m, n - ntile, tid);
	}
}

//
// Special matrix-vector multiplication kernel: for m < BLOCKSIZE_m (short and fat) cases.
//
// - Designed for architectures with atomicCAS, i.e., sm_11 and higher.
// - A block does multiple columns simultaneously.
// - Split into 2 separate device kernels.
// ---------------------------------------------------------------------
// |           several_cols_kernel3           | several_cols_kernel4 |
// ---------------------------------------------------------------------
// - Only valid if: 1) m * WORKSIZE_m <= BLOCKSIZE_m
//                  2) ORIGINAL_DESIGN: BLOCKSIZE_m * WORKSIZE_n <= max_shared_mem_per_block
// - On sm_11-sm_13 it uses own block_reduce kernel for block level reduction
//   and atomicCAS for grid level reduction.
// - Om >=sm_20 it uses atomicAdd for both block level and grid level reduction.

#define NSTEP ((BLOCKSIZE_m / WORKSIZE_m) * WORKSIZE_m)
#define NREST (BLOCKSIZE_m * WORKSIZE_m * WORKSIZE_n)
#define NTILE ((NREST / NSTEP) * NSTEP)

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void several_cols_kernel3(float alpha, const float* A, int lda, const float* x, float* y, float *x_s, int m, int tid)
{
	//
	// The bottom-left n-tiled part of the Matrix
	//
	int tidcol = tid / m;
	float sum = 0.0;
	for (int j = 0; j < NTILE; j += NSTEP)
	{
		__syncthreads();
		if (tid < NSTEP)
			x_s[tid]  = x[j];
		__syncthreads();
		if (tid < m * WORKSIZE_m)
		{
            #pragma unroll
			inner_loop_sf(NSTEP);
		}
	}
	__syncthreads();
	if (tid < NREST - NTILE)
		x_s[tid]  = x[NTILE];
	__syncthreads();
	if (tid < m * WORKSIZE_m)
	{
        #pragma unroll
		inner_loop_sf(NREST - NTILE);
	    atomicAdd(y - tidcol * m, alpha * sum);
	}
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void several_cols_kernel4(float alpha, const float* A, int lda, const float* x, float* y, float *x_s, int m, int nrest, int tid)
{
	//
	// The bottom-right rest of the Matrix
	//
	int tidcol = tid / m;
	int ntile = (nrest / NSTEP) * NSTEP;
	nrest -= ntile;
	float sum = 0.0;
	for (int j = 0; j < ntile; j += NSTEP)
	{
		__syncthreads();
		if (tid < NSTEP)
			x_s[tid]  = x[j];
		__syncthreads();
		if (tid < m * WORKSIZE_m)
		{
            #pragma unroll
			inner_loop_sf(NSTEP);
		}
	}
	__syncthreads();
	if (tid < nrest)
		x_s[tid]  = x[ntile];
	__syncthreads();
	if (tid < m * WORKSIZE_m)
	{
		int ntile2 = (nrest / WORKSIZE_m) * WORKSIZE_m;
		#pragma unroll 4
		inner_loop_sf(ntile2);
		if (tidcol < nrest - ntile2)
		{
			sum += A[0] * x_s[ntile2 + tidcol];
		}
	    atomicAdd(y - tidcol * m, alpha * sum);
	}
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m)
__global__ void sgemv_several_cols_kernel(float alpha, const float* A, int lda, const float* x, float* y, int m, int n, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int col = blockIdx.y * BLOCKSIZE_m * WORKSIZE_m * WORKSIZE_n;
	A += tid + col * lda;
	x += tid + col;
	y += tid;
	__shared__ float x_s[BLOCKSIZE_m];

	if (col != ntile)
	{
		several_cols_kernel3<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(alpha, A, lda, x, y, x_s, m, tid);
	}
	else
	{
		several_cols_kernel4<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(alpha, A, lda, x, y, x_s, m, n - ntile, tid);
	}
}

