#include "defines.h"
#include "sgemv_row_major_kernels.cu"
#include <cublas.h>

#if not defined (_TESLA) && not defined (_FERMI)
#error Requires an architechture flag (either _TESLA or _FERMI) to compile.
#endif

#define switchParameters alpha, A_d, lda, x_d, beta, y_d
#include "level2_templatemacros.h"

#ifdef _ATUNE

// When building atune
#define arguments sgemv_parameters
#define call_arguments sgemv_call_parameters
#define switchMacro1 workSizeSwitch_m_Large_n
#define switchMacro2 blockSizeSwitch_m
#define switchMacro3 blockSizeSwitch_m_sm
#define templatedec
#include "magma.h"

#else

// When building library
#define BLOCKSIZE BLOCKSIZE_m * BLOCKSIZE_n
#define blockSize_m BLOCKSIZE_m
#define blockSize_n BLOCKSIZE_n
#define workSize_m WORKSIZE_m
#define workSize_n WORKSIZE_n
#define arguments sgemv_arguments
#define call_arguments sgemv_call_arguments
#define switchMacro1 call_kernel_all
#define switchMacro2 call_kernel_all
#define switchMacro3 call_kernel_sm
#define templatedec template<int BLOCKSIZE_m, int BLOCKSIZE_n, int WORKSIZE_m, int WORKSIZE_n>
#include <cublas.h>

#endif

templatedec
static int sgemv_row_major_v1(arguments)
{		
	//
	// One-block-per-row version
	//

	// Determine size of grid
	int gridSize_m = m;
	int gridSize_n = 1;
		
	// Sizes of the tiled part of A
	int mtile = (m / blockSize_n) * blockSize_n;
	int ntile = (n / blockSize_n) * blockSize_n;
		
	// Kernel specifications
	dim3 dimGrid(gridSize_m, gridSize_n);
	dim3 dimBlock(blockSize_n);
		
	// Call nested switch macros for all templated values
	//printf("\nOne-block-per-row-%d: alpha=%f beta=%f m=%d n=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d work_m=%d work_n=%d\n", version, alpha, beta, m, n, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, workSize_m, workSize_n);
	sgemv_one_block_per_row_kernel<BLOCKSIZE, 1, 1><<< dimGrid, dimBlock >>>(switchParameters, m, n, mtile, ntile);
	return 0;
}

templatedec
static int sgemv_row_major_v2(arguments)
{		
	//
	// Several-blocks-per-row-2d version
	//
	
	// TODO: Make a kernel here that does the same as the short and fat version for sgemv_col_major
	//       See files in col major

	// Clear output vector if beta == 0.0, otherwise if beta != 1.0, scale output by beta
	if (beta == 0.0) cudaAssert( cudaMemset(y_d, 0, m * sizeof(float)) ); // make your own glas_memset !!
	//else (beta != 1.0) glas_sscal(m, beta, y_d, 1);

	// Determine size of grid
	int gridSize_m = (m + blockSize_m * workSize_m - 1) / (blockSize_m * workSize_m);
	int gridSize_n = 1;
	
	// Sizes of the tiled part of A
	int mtile = (m / (blockSize_m * workSize_m)) * (blockSize_m * workSize_m);
	int ntile = 0;
	
	// Kernel specifications
	dim3 dimGrid(gridSize_n, gridSize_m);
	dim3 dimBlock(blockSize_n, blockSize_m);
	int lds = n;
	int sizeSmem = blockSize_m * workSize_m * max(lds, blockSize_n) * sizeof(float);
	//int sizeSmem = (blockSize_m * workSize_m * max(lds, blockSize_n) + lds)* sizeof(float); // if X is in shared

	// Set 48 KB shared memory for this kernel
	//cudaFuncSetCacheConfig(sgemv_several_rows_per_block_x_shared_kernel<int, int, int>, cudaFuncCachePreferShared);
	// Call nested switch macros for all templated values
	//printf("\n2d-blocks-several-rows2-%d: m=%d n=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d block_n=%d work_m=%d work_n=%d lda=%d lds=%d sizeSmem=%d\n", version, m, n, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, blockSize_n, workSize_m, workSize_n, lda, lds, sizeSmem);
	switchMacro3(sgemv_xdim_block_per_row_2d_kernel, blockSize_m, blockSize_n, workSize_m, 1);
	return 0;
}

templatedec
static int sgemv_row_major_v3(arguments)
{		
	//
	// Several-blocks-several-rows version
	//

	// Clear output vector if beta == 0.0, otherwise if beta != 1.0, scale output by beta
	if (beta == 0.0) cudaAssert( cudaMemset(y_d, 0, m * sizeof(float)) ); // make your own glas_memset !!
	//else (beta != 1.0) glas_sscal(m, beta, y_d, 1);

	// Determine size of grid
	int gridSize_m = (m + workSize_m - 1) / workSize_m;
	int gridSize_n = (n + blockSize_n * workSize_n - 1) / (blockSize_n * workSize_n);
	
	// Sizes of the tiled part of A
	int mtile = (m / workSize_m) * workSize_m;
	int ntile = (n / (blockSize_n * workSize_n)) * (blockSize_n * workSize_n);
	
	// Kernel specifications
	dim3 dimGrid(gridSize_m, gridSize_n);
	dim3 dimBlock(blockSize_n);
	
	// Call nested switch macros for all templated values
	//printf("\nSeveral-blocks-several-rows-%d: m=%d n=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d block_n=%d work_m=%d work_n=%d\n", version, m, n, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, blockSize_n, workSize_m, workSize_n);
	switchMacro1(sgemv_several_blocks_several_rows_kernel, 1, BLOCKSIZE, workSize_m, workSize_n);
	return 0;
}

templatedec
static int sgemv_row_major_v4(arguments)
{		
	//
	// Several-blocks-per-row-2d version
	//
	
	// Clear output vector if beta == 0.0, otherwise if beta != 1.0, scale output by beta
	if (beta == 0.0) cudaAssert( cudaMemset(y_d, 0, m * sizeof(float)) ); // make your own glas_memset !!
	//else (beta != 1.0) glas_sscal(m, beta, y_d, 1);

	// Determine size of grid
	int gridSize_m = (m + blockSize_m * workSize_m - 1) / (blockSize_m * workSize_m);
	int gridSize_n = (n + blockSize_n * workSize_n - 1) / (blockSize_n * workSize_n);
	
	// Sizes of the tiled part of A
	int mtile = (m / (blockSize_m * workSize_m)) * (blockSize_m * workSize_m);
	int ntile = (n / (blockSize_n * workSize_n)) * (blockSize_n * workSize_n);
	
	// Kernel specifications
	dim3 dimGrid(gridSize_n, gridSize_m);
	dim3 dimBlock(blockSize_n, blockSize_m);
	
	// Call nested switch macros for all templated values
	//printf("\n2d-blocks-several-rows-%d: m=%d n=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d block_n=%d work_m=%d work_n=%d\n", version, m, n, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, blockSize_n, workSize_m, workSize_n);
	switchMacro2(sgemv_several_blocks_per_row_2d_kernel, blockSize_m, blockSize_n, workSize_m, workSize_n);
	return 0;
}

#ifdef _ATUNE

// When building atune
template <int BLOCKSIZE_n>
int sgemv_row_major(sgemv_parameters);

template <>
int sgemv_row_major<BLOCKSIZE>(sgemv_parameters)
{
	if (version == 1 && m <= 65536)
	{
		return sgemv_row_major_v1(sgemv_call_parameters);
	}
	else if (version == 2 && m <= 65536 * blockSize_m * workSize_m && n <= 65536 * blockSize_n && blockSize_m * workSize_m * max(n, blockSize_n) < (49152 / 2) / sizeof(float))
	{
		return sgemv_row_major_v2(sgemv_call_parameters);
	}
	else if (version == 3 && m <= 65536 * workSize_m && n <= 65536 * blockSize_n * workSize_n)
	{
		return sgemv_row_major_v3(sgemv_call_parameters);
	}
	else if (version == 4 && m <= 65536 * blockSize_m * workSize_m && n <= 65536 * blockSize_n * workSize_n)
	{
		return sgemv_row_major_v4(sgemv_call_parameters);
	}
	// return "not run" signal to auto-tuner
	return -1;
}

#else

//When building library
#include "atune_headers/sgemv_row_major.h"
//int atune_sgemv_col_major(sgemv_arguments);

/*
// See "sgemv_col_major.cu" for glas_sgemv() library function
int glas_sgemv(char trans, int m, int n, float alpha, float *A_d, int lda, float *x_d, int incx, float beta, float *y_d, int incy)
{
	if ((beta == 0.0 || beta == 1.0) && incx == 1 && incy == 1)
	{
		if (trans == 'N' || trans == 'n')
		{
			//return atune_sgemv_col_major(sgemv_call_arguments);
		}
		else if (trans == 'T' || trans == 't')
		{
			return atune_sgemv_row_major(sgemv_call_arguments);
		}
		else 
		{
			printf("Call to glas_sgemv: parameter 1 has an illegal value!\n"); exit(EXIT_FAILURE);
		}
	}
	else
	{
		printf("Call to sgemv: parameter 8, 9 or 11 has an unsupported value (%d, %f, %d)!\n", incx, beta, incy); exit(EXIT_FAILURE);
	}
	return -1;
}
*/
#endif

